<?php

namespace Imagine;

use resource;

class Canvas {

    public int $width;

    public int $height;

    public /*\resource|GdImage*/ $image; /* actual img resource */

    public Canvas $parent;

    public array $layers = [];

    public array $font = [];

    public Colors $colors;

    public function __construct($width = 0, $height = 0) {
        if ( $width && $height ) {
            $this->create($width, $height);
        }
    }

    public function __destruct() {
        $this->destroy();
    }

    public function destroy() : self
    {
        \is_resource($this->image) && \imagedestroy($this->image);

        return $this;
    }

    public function create(int $width, int $height) : self
    {
        $this->width = $width;
        $this->height = $height;

        $image = \imagecreatetruecolor($width, $height);

        $colors = new Colors($image);
        $colors['fore']  = empty($this->colors['fore'])  ? Colors::allocate($image, "000000")     : Colors::allocate($image, $this->colors['fore']);
        $colors['brush'] = empty($this->colors['brush']) ? Colors::allocate($image, "FFFFFF:0.0") : Colors::allocate($image, $this->colors['brush']);
        $colors['back']  = empty($this->colors['back'])  ? Colors::allocate($image, ":0.0")       : Colors::allocate($image, $this->colors['back']);
        $colors['text']  = empty($this->colors['text'])  ? Colors::allocate($image, "000000")     : Colors::allocate($image, $this->colors['text']);

        $this->colors = $colors;

        if ( empty($this->font['id']) ) {
            $this->font['id'] = 1;
        }

        \imagesavealpha($image, true);
        \imagealphablending($image, true);
        \imagefill($image, 0, 0, IMG_COLOR_TRANSPARENT);

        $this->image($image);

        return $this;
    }

    /* Getters / Setters */
    public function parent($set = null) {
        return $set === null ? $this->parent : $this->parent = $set;
    }

    public function width($set = null) {
        return $set === null ? $this->width : $this->width = $set;
    }

    public function height($set = null) {
        return $set === null ? $this->height : $this->height = $set;
    }

    public function & image($set = null) {
        if ( $set ) {
            $this->image = $set;
        };

        return $this->image;
    }

    public function layers($name = null, $replace = null) {
        return $name ? ( $replace ? $this->layers[$name] = $replace : $this->layers[$name] ) : $this->layers;
    }

    public function addLayer($name, $param = array()) : Layer
    {
        $param = $param + [
            'width'  => $this->width,
            'height' => $this->height,
            'parent' => $this
        ];

        return $this->layers[$name] = Layer::make($param)->onTop($this);
    }

    /*
     *  Assign a color to the current image resource
     *
     *  @param array $colors = [
     *      'color' => "aacc22:0.4"
     *      'background' => ':0.0"  # transparency
     *  ]
     *
     */
    public function colors($colors) : ? Colors
    {
        if ($colors && is_array($colors)) {
            $colorobj = Colors::make($this->image());

            foreach($colors as $key => $value) {
                $colorobj[$key] = $value;
            }

            return $colorobj;
        }

        return null;
    }

    public function measureText(string $text) : int
    {
        return Text::measure($text, $this->font['id']);
    }

    public function drawText(string $text, int $x, int $y, array $colors = []) : bool
    {
        return Text::draw($this->image(), $colors ?: $this->colors, $this->font['id'], $text, $x, $y);
    }

    public function polygonLabel($labels, $spacing, $x, $y, $size, $angle = 0) : array
    {
        $vertexcount = count($labels);
        $r = $size / 2;
        $x = $x;
        $y = $y;

        $coord = [];
        $points = Polygon::polygonCoord($vertexcount, $x, $y, $size, $angle);

        for ($i = 0; $i < $vertexcount; $i++) {

            $textsize = $this->measureText($labels[$i]);

            $tx = $points["x$i"];
            $ty = $points["y$i"];

            # processing X
            # we are in the center of the poly width
            if ( ! \bccomp($tx, $x) ) {
                $coord["x$i"] = $x - $textsize['width'] / 2;
            }
            else if ($tx > $x) {
                $coord["x$i"] = $tx + $spacing;
            }
            else {
                #echo $x . " = " . $tx . " <br>";
                $coord["x$i"] = $tx - $spacing - $textsize['width'];
            }

            # processing Y
            if ( ! \bccomp($ty, $y) ) {
                $coord["y$i"] = $y - $textsize['height'] / 2;
            }
            elseif ($ty > $y) {
                $coord["y$i"] = $ty + $spacing;
            }
            else {
                $coord["y$i"] = $ty - $textsize['height'] - $spacing;
            }
        }

       return $coord;
    }

    public function fillBorder($x, $y, $colors = null) : bool
    {
        return \imagefilltoborder($this->image(), $x, $y, $colors ? $colors['fore'] : $this->colors['fore'], $colors ? $colors['brush'] : $this->colors['brush']);
    }

    public function merge(resource $image, int $x = 0, int $y = 0, int $srcX = 0, int $srcY = 0) : self
    {
        Merge::simple($this->image(), $image, $x, $y, $srcX, $srcY);

        return $this;
    }

    public function grabImageResource(Canvas $src) : void
    {
        $this->destroy();
        $this->image($src->image());

        unset($src->image);
    }

    public function constrain(int $width, int $height, bool $immutable = true) : self
    {
        if ( ($this->width > $width) || ($this->height > $height) ) {
            if ($this->width > $this->height) {
                $height = $this->height * ( $width / $this->width );
            }
            else {
                $width = $this->width * ( $height / $this->height );
            }

            $new = ( new static() )->create($width, $height);
            $this->copyTo( $new );

            if ( ! $immutable ) {
                $this->grabImageResource($new);

                $this->width  = $width;
                $this->height = $height;
            }
            else {
                return $new;
            }
        }

        return $this;
    }

    public function crop(int $width, int $height, bool $centered = false, bool $immutable = true) : self
    {
        $x = $y = 0;

        if ($centered) {
            if ($this->width >= $this->height) {
                $r = $this->height / $height;
                $x = ($this->width / 2) - ($width * $r / 2);
            }
            else {
                $r = $this->width / $width;
                $y = ($this->height / 2) - ($height * $r / 2);
            }
        }

        $new = ( new static() )->create($width, $height);

        $this->copyTo($new, 0, 0, $x, $y, 0, 0, $this->width - ($x * 2), $this->height - ($y * 2));

        if ( ! $immutable ) {
            $this->grabImageResource($new);

            $this->width  = $width;
            $this->height = $height;
        }
        else {
            return $new;
        }


        return $this;
    }

    public function copyTo(Canvas $dest, int $dst_x = 0, int $dst_y = 0, int $src_x = 0, int $src_y = 0, int $dst_w = 0, int $dst_h = 0, int $src_w = 0, int $src_h = 0) : self
    {
        if ( ! \imagecopyresampled($dest->image(), $this->image(), $dst_x, $dst_y, $src_x, $src_y, $dst_w ?: $dest->width, $dst_h ?: $dest->height, $src_w ?: $this->width, $src_h ?: $this->height) ) {
            trigger_error('Given image was not resampled : an error occured', \E_USER_ERROR);
        }

        return $this;
    }

    public function drawLine(int $x1, int $y1, int $x2, int $y2, array $colors = []) : bool
    {
        return Draw::line($this->image(), $x1, $y1, $x2, $y2, $colors ? : $this->colors);
    }

    public function drawRectangle(int $x1, int $y1, int $width, int $height, array $colors = []) : bool
    {
        return Polygon::drawRectangle($this->image(), $x1, $y1, $x1 + $width, $y1 + $height, $colors ? $colors : $this->colors);
    }

    public function drawIrregularPolygon(array $points, int $x, int $y, $angle = 0, array $colors = []) : bool
    {
        return Polygon::drawIrregular($this->image(), $colors ? : $this->colors, $points, $x, $y, $angle);
    }

    public function drawPolygon(int $vertexcount, int $x, int $y, int $size, $angle = 0, $colors = []) : bool
    {
        return Polygon::draw($this->image(), $colors ? : $this->colors, $vertexcount, $x, $y, $size, $angle);
    }

    public function drawCircle(int $x, int $y, int $size, array $colors = []) : bool
    {
        return Polygon::drawEllipse($this->image(), $colors ? : $this->colors, $x, $y, $size, $size);
    }

    public function drawEllipse(int $x, int $y, int $width, int $height, array $colors = []) : bool
    {
        return Polygon::drawEllipse($this->image(), $colors ? : $this->colors, $x, $y, $width, $height);
    }

    public function drawTriangle(int $posX, int $posY, int $size, $angle = 0, array $colors = []) : bool
    {
        return Polygon::draw($this->image(), $colors ? : $this->colors, 3, $posX, $posY, $size, $angle);
    }

    public function drawQuadrilateral(int $posX, int $posY, int $size, $angle = 0, array $colors = []) : bool
    {
        return Polygon::draw($this->image(), $colors ? : $this->colors, 4, $posX, $posY, $size, $angle);
    }

    public function drawPentagon(int $posX, int $posY, $size, $angle = 0, array $colors = []) : bool
    {
        return Polygon::draw($this->image(), $colors ? : $this->colors, 5, $posX, $posY, $size, $angle);
    }

    public function drawHexagon(int $posX, int $posY, int $size, $angle = 0, array $colors = []) : bool
    {
        return Polygon::draw($this->image(), $colors ? : $this->colors, 6, $posX, $posY, $size, $angle);
    }

    public function drawHeptagon(int $posX, int $posY, int $size, $angle = 0, array $colors = []) : bool
    {
        return Polygon::draw($this->image(), $colors ? : $this->colors, 7, $posX, $posY, $size, $angle);
    }

    public function drawOctagon(int $posX, int $posY, int $size, $angle = 0, array $colors = []) : bool
    {
        return Polygon::draw($this->image(), $colors ? : $this->colors, 8, $posX, $posY, $size, $angle);
    }

    public function drawNonagon(int $posX, int $posY, int $size, $angle = 0, array $colors = []) : bool
    {
        return Polygon::draw($this->image(), $colors ? : $this->colors, 9, $posX, $posY, $size, $angle);
    }

    public function drawDecagon(int $posX, int $posY, int $size, $angle = 0, array $colors = []) : bool
    {
        return Polygon::draw($this->image(), $colors ? : $this->colors, 10, $posX, $posY, $size, $angle);
    }
}
